#!/usr/bin/env python3
"""
Flyff HP Total Node (Final Combined Version)
 - Polls the API at http://127.0.0.1:5000/data
 - Outputs only the "hp_total" value as a string
 - Uses color (126, 36, 57) for its output port
 - Displays "hp_total" in a text field labeled "Value"
 - Retrieves the port with self.outputs().get('value')
"""

import time
import requests
import traceback
from OdenGraphQt import BaseNode

class FlyffHPTotalNode(BaseNode):
    __identifier__ = 'bunny-lab.io.flyff_hp_total_node'
    NODE_NAME = 'Flyff - HP Total'

    def __init__(self):
        super(FlyffHPTotalNode, self).__init__()

        # 1) Text input property named "value" for UI display
        self.add_text_input('value', 'Value', text='N/A')

        # 2) Output port also named "value"
        self.add_output('value', color=(126, 36, 57))

        self._api_down = True
        self._last_api_attempt = 0.0
        self._retry_interval = 5.0
        self._last_error_printed = 0.0

        self.set_name("Flyff - HP Total (API Disconnected)")

    def process_input(self):
        current_time = time.time()
        if self._api_down and (current_time - self._last_api_attempt < self._retry_interval):
            return

        self._last_api_attempt = current_time

        try:
            response = requests.get("http://127.0.0.1:5000/data", timeout=1)
            status_code = response.status_code
            print(f"[DEBUG] FlyffHPTotalNode: HTTP Status Code = {status_code}")

            if status_code == 200:
                try:
                    data = response.json() or {}
                except ValueError:
                    data = {}

                if isinstance(data, list):
                    data = {}

                self._api_down = False
                self.set_name("Flyff - HP Total (API Connected)")

                new_value = data.get("hp_total", "N/A")
                print(f"[DEBUG] FlyffHPTotalNode: hp_total = {new_value}")

                new_value_str = str(new_value)
                self.set_property('value', new_value_str)
                self.transmit_data(new_value_str)

            else:
                self._handle_api_error(f"HTTP {status_code} from FlyffHPTotalNode")
                self._api_down = True

        except Exception as e:
            tb = traceback.format_exc()
            self._handle_api_error(f"Exception in FlyffHPTotalNode: {e}\nTraceback:\n{tb}")
            self._api_down = True

    def transmit_data(self, data):
        output_port = self.outputs().get('value')
        if output_port and output_port.connected_ports():
            for connected_port in output_port.connected_ports():
                connected_node = connected_port.node()
                if hasattr(connected_node, 'receive_data'):
                    try:
                        connected_node.receive_data(data, source_port_name='value')
                    except Exception as e:
                        print(f"[ERROR] Error transmitting data to {connected_node}: {e}")

    def _handle_api_error(self, msg):
        current_time = time.time()
        if (current_time - self._last_error_printed) >= self._retry_interval:
            print(f"[ERROR] {msg}")
            self._last_error_printed = current_time

        self.set_name("Flyff - HP Total (API Disconnected)")
